from __future__ import annotations

import typing
from abc import ABCMeta, abstractmethod
from itertools import islice

from dynamicprompts.commands import Command
from dynamicprompts.parser.config import ParserConfig, default_parser_config
from dynamicprompts.parser.parse import parse
from dynamicprompts.utils import squash_whitespace
from dynamicprompts.wildcardmanager import WildcardManager


class Sampler(metaclass=ABCMeta):
    def __init__(
        self,
        *,
        wildcard_manager: WildcardManager,
        parser_config: ParserConfig = default_parser_config,
        ignore_whitespace: bool = False,
    ):
        self._wildcard_manager = wildcard_manager
        self._ignore_whitespace = ignore_whitespace
        self._parser_config = parser_config

    @abstractmethod
    def generator_from_command(
        self,
        command: Command,
    ) -> typing.Generator[str, None, None]:
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement generator_from_command",
        )

    def generate_prompts(
        self,
        prompt: str | Command,
        num_prompts: int | None = None,
    ) -> typing.Iterable[str]:
        """
        Generate prompts from a prompt template.

        :param prompt: The prompt template to generate prompts from.
        :param num_prompts: How many prompts to generate (at most). If None, generate all possible prompts.
        """
        if not prompt:
            return []
        command: Command
        if isinstance(prompt, str):
            command = parse(prompt, parser_config=self._parser_config)
        elif isinstance(prompt, Command):
            command = prompt
        else:
            raise TypeError(f"Expected prompt to be str or Command, got {type(prompt)}")
        gen = self.generator_from_command(command)

        if self._ignore_whitespace:
            gen = (squash_whitespace(p) for p in gen)

        if num_prompts is None:
            return gen
        return islice(gen, num_prompts)



from __future__ import annotations

import logging
from typing import Iterable

from dynamicprompts import constants
from dynamicprompts.generators.promptgenerator import PromptGenerator
from dynamicprompts.parser.parse import default_parser_config
from dynamicprompts.samplers.combinatorial import CombinatorialSampler
from dynamicprompts.wildcardmanager import WildcardManager

logger = logging.getLogger(__name__)


class CombinatorialPromptGenerator(PromptGenerator):
    def __init__(
        self,
        wildcard_manager: WildcardManager,
        ignore_whitespace: bool = False,
    ) -> None:
        self._wildcard_manager = wildcard_manager
        self._sampler = CombinatorialSampler(
            wildcard_manager=wildcard_manager,
            ignore_whitespace=ignore_whitespace,
            parser_config=default_parser_config,
        )

    def generate(  # type: ignore[override]
        self,
        template: str | None,
        max_prompts: int | None = constants.MAX_IMAGES,
    ) -> Iterable[str]:
        if template is None or len(template) == 0:
            return [""]
        prompts = self._sampler.generate_prompts(template, max_prompts)

        return prompts



from __future__ import annotations

import re

DEFAULT_ENCODING = "utf-8"
DEFAULT_COMBO_JOINER = ","
MAX_IMAGES = 1000
WILDCARD_SUFFIX = "txt"
WILDCARD_RE = re.compile(r"__(.*?)__")
COMBINATIONS_RE = re.compile(r"\{([^{}]*)}")



from __future__ import annotations


def squash_whitespace(s: str) -> str:
    return " ".join(s.split())


def is_empty_line(line: str | None) -> bool:
    return line is None or line.strip() == "" or line.strip().startswith("#")



from __future__ import annotations

import dataclasses

from dynamicprompts.commands import Command


@dataclasses.dataclass
class LiteralCommand(Command):
    literal: str

    def __add__(self, other):
        if isinstance(other, LiteralCommand):
            return LiteralCommand(f"{self.literal} {other.literal}")
        raise TypeError(f"Cannot concatenate LiteralCommand with {other}")



from __future__ import annotations

import logging
import typing
from random import Random

from dynamicprompts.commands import (
    Command,
    LiteralCommand,
    SequenceCommand,
    VariantCommand,
    WildcardCommand,
)
from dynamicprompts.parser.config import ParserConfig, default_parser_config
from dynamicprompts.samplers.base import Sampler
from dynamicprompts.wildcardmanager import WildcardManager

logger = logging.getLogger(__name__)

DEFAULT_RANDOM = Random()


def _get_random_variant(
    generator: Sampler,
    variant_command: VariantCommand,
    random: Random,
) -> typing.Generator[str, None, None]:
    if len(variant_command.values) == 0:
        return
    elif len(variant_command.values) == 1:
        yield from generator.generator_from_command(variant_command.values[0])
    else:
        while True:
            num_choices = random.randint(
                variant_command.min_bound,
                variant_command.max_bound,
            )
            selected_commands = random.choices(
                variant_command.values,
                weights=variant_command.weights,
                k=num_choices,
            )
            sub_generators = [
                generator.generator_from_command(c) for c in selected_commands
            ]
            yield variant_command.separator.join(
                next(subgen) for subgen in sub_generators
            )


class RandomSampler(Sampler):
    def __init__(
        self,
        *,
        wildcard_manager: WildcardManager,
        ignore_whitespace: bool = False,
        rand: Random = DEFAULT_RANDOM,
        parser_config: ParserConfig = default_parser_config,
    ):
        super().__init__(
            wildcard_manager=wildcard_manager,
            ignore_whitespace=ignore_whitespace,
            parser_config=parser_config,
        )
        self._random = rand

    def generator_from_command(
        self,
        command: Command,
    ) -> typing.Generator[str, None, None]:
        if isinstance(command, LiteralCommand):
            while True:
                yield command.literal
        elif isinstance(command, SequenceCommand):
            sub_generators = [self.generator_from_command(c) for c in command.tokens]
            while True:
                yield command.separator.join(next(subgen) for subgen in sub_generators)
        elif isinstance(command, VariantCommand):
            yield from _get_random_variant(self, command, self._random)
        elif isinstance(command, WildcardCommand):
            values = self._wildcard_manager.get_all_values(command.wildcard)
            while True:
                if len(values) == 0:
                    logger.warning(f"No values found for wildcard {command.wildcard}")
                    yield f"__{command.wildcard}__"
                else:
                    value = self._random.choice(values)
                    # Parse and generate prompts from wildcard value
                    yield from self.generate_prompts(value, 1)
        else:
            raise NotImplementedError(
                f"{self.__class__.__qualname__}: Not implemented: {command}",
            )



from __future__ import annotations

import typing

from dynamicprompts.commands import (
    Command,
    LiteralCommand,
    SequenceCommand,
    VariantCommand,
    WildcardCommand,
)
from dynamicprompts.samplers.base import Sampler


def _get_combinatorial_sequence(
    generator: Sampler,
    tokens: list[Command],
    *,
    separator: str,
) -> typing.Iterable[str]:
    if not tokens:
        yield ""
        return
    token = tokens[0]
    for prompt in generator.generator_from_command(token):
        for next_prompts in _get_combinatorial_sequence(
            generator,
            tokens[1:],
            separator=separator,
        ):
            res = prompt + separator + next_prompts
            yield res


def _dedupe(arr: list[str]) -> tuple[str, ...]:
    seen: set[str] = set()
    result: list[str] = []
    for item in arr:
        if item not in seen:
            seen.add(item)
            result.append(item)
    return tuple(result)


def _combo_to_prompt(
    generator: Sampler,
    combo: list[Command],
) -> typing.Iterable[list[str]]:
    if len(combo) == 0:
        yield []
        return
    c_1, c_rest = combo[0], combo[1:]

    for p in generator.generator_from_command(c_1):
        for rest_prompt in _combo_to_prompt(generator, c_rest):
            if rest_prompt:
                yield [p] + rest_prompt
            else:
                yield [p]


def _get_combinatorial_variant(
    generator: Sampler,
    variant_command: VariantCommand,
) -> typing.Iterable[str]:
    if len(variant_command.variants) == 0:
        return

    seen = set()

    for bound in range(variant_command.min_bound, variant_command.max_bound + 1):
        for combo in variant_command.get_value_combinations(bound):
            for prompt_arr in _combo_to_prompt(generator, combo):
                deduped_arr = _dedupe(prompt_arr)
                correct_size = len(deduped_arr) == bound
                if correct_size and deduped_arr not in seen:
                    seen.add(deduped_arr)
                    yield variant_command.separator.join(deduped_arr)


class CombinatorialSampler(Sampler):
    def generator_from_command(
        self,
        command: Command,
    ) -> typing.Generator[str, None, None]:
        if isinstance(command, LiteralCommand):
            yield command.literal
        elif isinstance(command, SequenceCommand):
            yield from _get_combinatorial_sequence(
                self,
                command.tokens,
                separator=command.separator,
            )
        elif isinstance(command, VariantCommand):
            yield from _get_combinatorial_variant(self, command)
        elif isinstance(command, WildcardCommand):
            for val in self._wildcard_manager.get_all_values(command.wildcard):
                # Parse and generate prompts from wildcard value
                yield from self.generate_prompts(val)
        else:
            raise NotImplementedError(
                f"{self.__class__.__name__} does not support {command.__class__.__name__}",
            )
